"""Update dependencies stated in the Pipfile.lock for the given repo or repositories."""

from .update import update

__name__ = 'kebechet'
__version__ = '1.0.0'
__author__ = 'Fridolin Pokorny <fridolin.pokorny@gmail.com>'
